print('Script start')
import csv
import math
import cmath
import numpy as np
from pyomo.environ import *
import matplotlib.pyplot as plt

print('Libraries imported')

#Import bus data
with open('data/bus.csv', newline='') as csvfile:
    bus = np.array(list(csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)))

#Load profile
with open('data/24hLoad.csv', newline='') as csvfile:
    load24h = np.array(list(csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)))

#Irradiation
with open('data/24hIrradiation.csv', newline='') as csvfile:
    irrad = np.array(list(csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)))

#yabs
with open('data/yabs.csv', newline='') as csvfile:
    yabs = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
#yang
with open('data/yang.csv', newline='') as csvfile:
    yang = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
#Ji3
with open('data/Ji3.csv', newline='') as csvfile:
    ji3 = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
#Ji4
with open('data/Ji4.csv', newline='') as csvfile:
    ji4 = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
model = ConcreteModel()
syslen = 33#33
timestart = 1#1
timestop = 24#24
pv = 30#30
slack = 1
#day = [1, 2, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]
#night = [n for n in range(3,9)]
day = [d for d in range(9,17)]
night = [1,2,3,4,5,6,7,8,17, 18, 19, 20, 21, 22, 23, 24]
#SETS--------------------------------------------------------------------------------------------------------------------
def i_init(model):
    return range(1,syslen+1)
def t_init(model):
    return range(timestart,timestop+1)
model.i = Set(within=Integers, initialize=i_init, doc='Buses')
model.t = Set(within=Integers, initialize=t_init, doc='Time of day')
print('Sets declared')
#SCALARS----------------------------------------------------------------------------------------------------------------------
v_min = 0.95
v_max = 1.05
kup = 1.2816
kdown = -1.2816
deghat = 0.1
print('Scalars declared')
#PARAMETERS--------------------------------------------------------------------------------------------------------------------
def P_time_init(model, t):
    return float(load24h[t-1])
model.P_time = Param(model.t, initialize=P_time_init, doc='Percent of power in time')
def p_peak_init(model, i):
    return float(bus[i-1,2])
model.p_peak = Param(model.i, initialize=p_peak_init, doc='Peak power at bus')
def q_peak_init(model, i):
    return float(bus[i-1,3])
model.q_peak = Param(model.i, initialize=q_peak_init, doc='Peak reactive power at bus')
def Pl_init(model, i, t):
    return model.P_time[t] * model.p_peak[i] 
model.Pl = Param(model.i, model.t, initialize=Pl_init, doc='Active load')
def Ql_init(model, i, t):
    return model.P_time[t] * model.q_peak[i] 
model.Ql = Param(model.i, model.t, initialize=Ql_init, doc='Reactive load')


def sigmaq_init(model, i, t):
    return (abs(model.Ql[i,t])/3)*deghat
def sigmap_init(model, i, t):
    return (model.Pl[i,t]/3)*deghat

model.sigmaq = Param(model.i, model.t, initialize=sigmaq_init)
model.sigmap = Param(model.i, model.t, initialize=sigmap_init)

def sigmav_init(model, i, t):
    return math.sqrt(sum(((abs(ji3[i-1,j-1])*model.sigmap[i,t])**2+(abs(ji4[i-1,j-1])*model.sigmaq[i,t])**2)for j in range(1,syslen+1)))
model.sigmav = Param(model.i, model.t, initialize=sigmav_init)
print('Parameters declared')
#VARIABLES--------------------------------------------------------------------------------------------------------------------
model.delta = Var(model.i, model.t, doc='Voltage angle')
model.pn = Var(model.i, model.t, doc='Net active power')
model.qn = Var(model.i, model.t, doc='Net reactive power')
model.qg = Var(model.i, model.t, doc='Reactive power generation')
model.pg = Var(model.i, model.t, doc='Active power generation')

model.y = Var(model.i, model.t, bounds=(0,None), doc='Objective function')
model.v = Var(model.i, model.t, bounds=(0,None), doc='Voltage magnitude')
model.loss = Var(model.t, bounds=(0,None), doc='Losses')
model.v_slack = Var(model.t, bounds=(0,None), doc='Slack voltage')

model.k = Var(model.t, within=Integers, bounds=(None,16), doc='Tap changer tap')
print('Variables declared')
#EQUATIONS--------------------------------------------------------------------------------------------------------------------
def y1_eq(model, i, t):
    return model.y[i,t] >= model.v[i,t] - 1
model.y1 = Constraint(model.i, model.t, rule=y1_eq)

def y2_eq(model, i, t):
    return model.y[i,t] >= 1 - model.v[i,t]
model.y2 = Constraint(model.i, model.t, rule=y2_eq)

def vslack_eq(model, t):
    return model.v[slack,t] == model.v_slack[t]
model.vslack = Constraint(model.t, rule=vslack_eq)

def tap_pos_eq(model, t):
    return model.v_slack[t] == 1+(model.k[t]-8)*0.00625
model.tap_pos = Constraint(model.t, rule=tap_pos_eq)

def deltaslack_eq(model, t):
    return model.delta[slack,t] == 0
model.deltaslack = Constraint(model.t, rule=deltaslack_eq)

def voltage_min_limit_eq(model, i, t):
    return model.v[i,t] >= v_min - model.sigmav[i,t]*kdown
model.voltage_min_limit = Constraint(model.i, model.t, rule=voltage_min_limit_eq)

def voltage_max_limit_eq(model, i, t):
    return model.v[i,t] <= v_max - model.sigmav[i,t]*kup
model.voltage_max_limit = Constraint(model.i, model.t, rule=voltage_max_limit_eq)

def activelf_eq(model, i, t):
    return model.pn[i,t] == sum((yabs[i-1,j-1]*(cos(yang[i-1,j-1])+cos(yang[i-1,j-1])*(model.v[i,t]-1)+cos(yang[i-1,j-1])*(model.v[j,t]-1)-sin(yang[i-1,j-1])*(model.delta[j,t])
                                                +sin(yang[i-1,j-1])*(model.delta[i,t])))for j in range(1,syslen+1))
model.activelf = Constraint(model.i, model.t, rule=activelf_eq)

def reactivelf_eq(model, i, t):
    return model.qn[i,t] == sum((yabs[i-1,j-1]*(cos(yang[i-1,j-1])*(model.delta[i,t])-(sin(yang[i-1,j-1])*(model.v[i,t]+model.v[j,t]-1)+cos(yang[i-1,j-1])
                                                                                       *(model.delta[j,t]))))for j in range(1,syslen+1))
model.reactivelf = Constraint(model.i, model.t, rule=reactivelf_eq)

def kclre1_eq(model, i, t):
    return model.pn[i,t] == model.pg[i,t] - model.Pl[i,t]
model.kclre1 = Constraint(model.i, model.t, rule=kclre1_eq)

def kclim1_eq(model, i, t):
    return model.qn[i,t] == model.qg[i,t] - model.Ql[i,t]
model.kclim1 = Constraint(model.i, model.t, rule=kclim1_eq)

def eq1_eq(model, i, t):
    if (i!=slack):
        return model.pg[i,t] == 0
    return Constraint.Skip
model.eq1 = Constraint(model.i, model.t, rule=eq1_eq)

def eq2_eq(model, i, t):
    if (i!=slack):
        return model.qg[i,t] == 0
    return Constraint.Skip
model.eq2 = Constraint(model.i, model.t, rule=eq2_eq)

def pslack_eq(model, t):
    return model.pg[slack,t] >= 0
model.pslack = Constraint(model.t, rule=pslack_eq)

def qslack_eq(model, t):
    return model.qg[slack,t] >= 0
model.qslack = Constraint(model.t, rule=qslack_eq)

def l_eq(model, t):
    return model.loss[t] == sum((model.pg[i,t]-model.Pl[i,t])for i in model.i)
model.l = Constraint(model.t, rule=l_eq)

print('Equations declared')
#OBJECTIVE--------------------------------------------------------------------------------------------------------------------
def objective_function(model):
    return sum(model.y[i,t] for i in model.i for t in model.t)
model.objective = Objective(rule=objective_function, sense=minimize, doc='Minimize voltage difference')
print('Objective declared')

print('Done constructing model!')
def pyomo_postprocess(options=None, instance=None, results=None):
    print('Pyomo printing results to file...')
    model.pprint(filename='results/no_PV/results.txt')
if __name__ == '__main__':
    # This emulates what the pyomo command-line tools does
    from pyomo.opt import SolverFactory
    import pyomo.environ
    opt = SolverFactory("gurobi")
    results = opt.solve(model)#, tee=True)
    #sends results to stdout
    results.write()
    #print("\nDisplaying Solution\n" + '-'*60)
    #pyomo_postprocess(None, model, results)

print('Postprocessing results')
v = np.zeros((syslen,timestop))
delta = np.zeros((syslen,timestop))
V_max = np.zeros((syslen,timestop))
V_min = np.zeros((syslen,timestop))
pg = np.zeros((syslen,timestop))
qg = np.zeros((syslen,timestop))
loss = np.zeros(timestop)
Pl = np.zeros(timestop)
Ql = np.zeros(timestop)
for t in model.t:
    for i in model.i:
        v[i-1,t-1] = value(model.v[i,t])
        delta[i-1,t-1] = value(model.delta[i,t])
        pg[i-1,t-1] = value(model.pg[i,t])
        qg[i-1,t-1] = value(model.qg[i,t])
        V_max[i-1,t-1] = v_max - model.sigmav[i,t]*kup
        V_min[i-1,t-1] = v_min - model.sigmav[i,t]*kdown
        
    loss[t-1] = value(model.loss[t])
    Pl[t-1] = sum(model.Pl[i,t] for i in model.i)
    Ql[t-1] = sum(model.Ql[i,t] for i in model.i)
t = [h for h in range(1,timestop+1)]
i = [ j for j in range(1,syslen+1)]

print('Plotting results...')
plt.rcParams["figure.figsize"] = (16,9)
plt.rcParams["axes.labelsize"] = 'x-large'
#GJENGANGERE------------------------------------------------------------------
plt.plot(i,v[:,19],label='Peak load hour')
plt.plot(i,v[:,11], label='Peak PV hour')
plt.legend(loc='best')
plt.xlabel('Bus number')
plt.ylabel('Voltage (P.U)')
plt.grid()
plt.savefig('results/no_PV/new/V_peaks.pdf', bbox_inches='tight')
plt.close()


#GJENGANGERE------------------------------------------------------------------
plt.plot(i,v[:,19],label='Voltage at peak load')
plt.plot(i,v[:,4], label='Voltage at min load')
plt.plot(i, V_max[:,19], dashes=[6,2], label='Upper voltage constraint')
plt.plot(i, V_min[:,19], dashes=[6,2], label='Lower voltage constraint')
plt.legend(loc='best')
plt.xlabel('Bus number')
plt.ylabel('Voltage (P.U)')
plt.grid()
plt.savefig('results/no_PV/new/Voltage_limits.pdf', bbox_inches='tight')
plt.close()

print('All done!')
