print('Script start')
import csv
import math
import cmath
import numpy as np
from pyomo.environ import *
import matplotlib.pyplot as plt

print('Libraries imported')

#Import bus data
with open('data/bus.csv', newline='') as csvfile:
    bus = np.array(list(csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)))

#Load profile
with open('data/24hLoad.csv', newline='') as csvfile:
    load24h = np.array(list(csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)))

#Irradiation
with open('data/24hIrradiation.csv', newline='') as csvfile:
    irrad = np.array(list(csv.reader(csvfile, quoting=csv.QUOTE_NONNUMERIC)))

#yabs
with open('data/yabs.csv', newline='') as csvfile:
    yabs = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
#yang
with open('data/yang.csv', newline='') as csvfile:
    yang = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
#Ji3
with open('data/Ji3.csv', newline='') as csvfile:
    ji3 = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
#Ji4
with open('data/Ji4.csv', newline='') as csvfile:
    ji4 = np.array(list(csv.reader(csvfile, delimiter=';', quoting=csv.QUOTE_NONNUMERIC)))
model = ConcreteModel()
syslen = 33#33
timestart = 1#1
timestop = 24#24
pv = 30#30
slack = 1
#day = [1, 2, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]
#night = [n for n in range(3,9)]
day = [d for d in range(9,17)]
night = [1,2,3,4,5,6,7,8,17, 18, 19, 20, 21, 22, 23, 24]
#SETS--------------------------------------------------------------------------------------------------------------------
def i_init(model):
    return range(1,syslen+1)
def t_init(model):
    return range(timestart,timestop+1)
model.i = Set(within=Integers, initialize=i_init, doc='Buses')
model.t = Set(within=Integers, initialize=t_init, doc='Time of day')
print('Sets declared')
#SCALARS----------------------------------------------------------------------------------------------------------------------
v_min = 0.95
v_max = 1.05
kup = 1.2816
kdown = -1.2816
deghat = 1
PVscale = 1
INVscale = 1
print('Scalars declared')
#PARAMETERS--------------------------------------------------------------------------------------------------------------------
def P_time_init(model, t):
    return float(load24h[t-1])
model.P_time = Param(model.t, initialize=P_time_init, doc='Percent of power in time')
def p_peak_init(model, i):
    return float(bus[i-1,2])
model.p_peak = Param(model.i, initialize=p_peak_init, doc='Peak power at bus')
def q_peak_init(model, i):
    return float(bus[i-1,3])
model.q_peak = Param(model.i, initialize=q_peak_init, doc='Peak reactive power at bus')
def Pl_init(model, i, t):
    return model.P_time[t] * model.p_peak[i] 
model.Pl = Param(model.i, model.t, initialize=Pl_init, doc='Active load')
def Ql_init(model, i, t):
    return model.P_time[t] * model.q_peak[i] 
model.Ql = Param(model.i, model.t, initialize=Ql_init, doc='Reactive load')
def solar_peak_init(model, i):
    if i==pv:
        return 1*PVscale
    return 0
model.solar_peak = Param(model.i, initialize=solar_peak_init, doc='Peak solar gen at bus in MW')
def Psolar_time_init(model, t):
    return float(irrad[t-1])
model.Psolar_time = Param(model.t, initialize=Psolar_time_init, doc='Percent of power in time')
#model.s_peak = Param(model.t)
def p_pvmppt_init(model, i, t):
    return model.Psolar_time[t]*model.solar_peak[i]
model.p_pvmppt = Param(model.i, model.t, initialize=p_pvmppt_init, doc='PV gen')
#model.mppt = Param(model.t)
#model.eta_inv = Param(model.i)
def inv_cap_init(model, i):
    if i==pv:
        return 1.054*INVscale
model.inv_cap = Param(model.i, initialize=inv_cap_init, doc='Inverter capacity')
def qday_init(model, i):
    if i==pv:
        return 0.3266*model.inv_cap[i]
def qnight_init(model, i):
    if i==pv:
        return 1
def pday_init(model, i):
    if i==pv:
        return 1
def pnight_init(model, i):
    if i==pv:
        return 0.3266*model.inv_cap[i]
model.qday = Param(model.i, initialize=qday_init)
model.qnight = Param(model.i, initialize=qnight_init)
model.pday = Param(model.i, initialize=pday_init)
model.pnight = Param(model.i, initialize=pnight_init)

def sigmagl_init(model, i, t):
    return math.sqrt(((model.p_pvmppt[i,t]/3)*deghat)**2+((model.Pl[i,t]/3)*deghat)**2)
def sigmaq_init(model, i, t):
    return (abs(model.Ql[i,t])/3)*deghat
def sigmap_init(model, i, t):
    return (model.Pl[i,t]/3)*deghat
def sigmag_init(model, i, t):
    return (model.p_pvmppt[i,t]/3)*deghat
model.sigmagl = Param(model.i, model.t, initialize=sigmagl_init)
model.sigmaq = Param(model.i, model.t, initialize=sigmaq_init)
model.sigmap = Param(model.i, model.t, initialize=sigmap_init)
model.sigmag = Param(model.i, model.t, initialize=sigmag_init)

def sigmav_init(model, i, t):
    return math.sqrt(sum(((abs(ji3[i-1,j-1])*model.sigmagl[i,t])**2+(abs(ji4[i-1,j-1])*model.sigmaq[i,t])**2)for j in range(1,syslen+1)))
model.sigmav = Param(model.i, model.t, initialize=sigmav_init)
print('Parameters declared')
#VARIABLES--------------------------------------------------------------------------------------------------------------------
model.delta = Var(model.i, model.t, doc='Voltage angle')
model.pn = Var(model.i, model.t, doc='Net active power')
model.qn = Var(model.i, model.t, doc='Net reactive power')
model.qg = Var(model.i, model.t, doc='Reactive power generation')
model.pg = Var(model.i, model.t, doc='Active power generation')

model.y = Var(model.i, model.t, bounds=(0,None), doc='Objective function')
model.v = Var(model.i, model.t, bounds=(0,None), doc='Voltage magnitude')
model.loss = Var(model.t, bounds=(0,None), doc='Losses')
model.v_slack = Var(model.t, bounds=(0,None), doc='Slack voltage')

model.k = Var(model.t, within=Integers, bounds=(None,16), doc='Tap changer tap')
model.Pdel = Var(model.i, model.t, doc='Juksevariabel P')
model.Qdel = Var(model.i, model.t, doc='Juksevariabel Q')
print('Variables declared')
#EQUATIONS--------------------------------------------------------------------------------------------------------------------
def y1_eq(model, i, t):
    return model.y[i,t] >= model.v[i,t] - 1
model.y1 = Constraint(model.i, model.t, rule=y1_eq)

def y2_eq(model, i, t):
    return model.y[i,t] >= 1 - model.v[i,t]
model.y2 = Constraint(model.i, model.t, rule=y2_eq)

def vslack_eq(model, t):
    return model.v[slack,t] == model.v_slack[t]
model.vslack = Constraint(model.t, rule=vslack_eq)

def tap_pos_eq(model, t):
    return model.v_slack[t] == 1+(model.k[t]-8)*0.00625
model.tap_pos = Constraint(model.t, rule=tap_pos_eq)

def deltaslack_eq(model, t):
    return model.delta[slack,t] == 0
model.deltaslack = Constraint(model.t, rule=deltaslack_eq)

def voltage_min_limit_eq(model, i, t):
    return model.v[i,t] >= v_min - model.sigmav[i,t]*kdown
model.voltage_min_limit = Constraint(model.i, model.t, rule=voltage_min_limit_eq)

def voltage_max_limit_eq(model, i, t):
    return model.v[i,t] <= v_max - model.sigmav[i,t]*kup
model.voltage_max_limit = Constraint(model.i, model.t, rule=voltage_max_limit_eq)

def activelf_eq(model, i, t):
    return model.pn[i,t] == sum((yabs[i-1,j-1]*(cos(yang[i-1,j-1])+cos(yang[i-1,j-1])*(model.v[i,t]-1)+cos(yang[i-1,j-1])*(model.v[j,t]-1)-sin(yang[i-1,j-1])*(model.delta[j,t])
                                                +sin(yang[i-1,j-1])*(model.delta[i,t])))for j in range(1,syslen+1))
model.activelf = Constraint(model.i, model.t, rule=activelf_eq)

def reactivelf_eq(model, i, t):
    return model.qn[i,t] == sum((yabs[i-1,j-1]*(cos(yang[i-1,j-1])*(model.delta[i,t])-(sin(yang[i-1,j-1])*(model.v[i,t]+model.v[j,t]-1)+cos(yang[i-1,j-1])
                                                                                       *(model.delta[j,t]))))for j in range(1,syslen+1))
model.reactivelf = Constraint(model.i, model.t, rule=reactivelf_eq)

def kclre1_eq(model, i, t):
    return model.pn[i,t] == model.pg[i,t] - model.Pl[i,t]
model.kclre1 = Constraint(model.i, model.t, rule=kclre1_eq)

def kclim1_eq(model, i, t):
    return model.qn[i,t] == model.qg[i,t] - model.Ql[i,t]
model.kclim1 = Constraint(model.i, model.t, rule=kclim1_eq)

def eq1_eq(model, i, t):
    if (i!=pv and i!=slack):
        return model.pg[i,t] == 0
    return Constraint.Skip
model.eq1 = Constraint(model.i, model.t, rule=eq1_eq)

def eq2_eq(model, i, t):
    if (i!=pv and i!=slack):
        return model.qg[i,t] == 0
    return Constraint.Skip
model.eq2 = Constraint(model.i, model.t, rule=eq2_eq)

def inverter_cap1_eq(model, t):
    if t in day:#model.p_pvmppt[pv,t]!=0:
        return -model.pg[pv,t]+((model.pday[pv]-model.inv_cap[pv])/model.qday[pv])*model.qg[pv,t]+model.inv_cap[pv]>=model.sigmag[pv,t]*kup
    return Constraint.Skip
model.inverter_cap1 = Constraint(model.t, rule=inverter_cap1_eq)

def inverter_cap2_eq(model, t):
    if t in day:#model.p_pvmppt[pv,t]!=0:
        return model.inv_cap[pv]+((model.pday[pv]-model.inv_cap[pv])/model.qday[pv])*model.qg[pv,t]-model.pg[pv,t]>=model.sigmag[pv,t]*kup
    return Constraint.Skip
model.inverter_cap2 = Constraint(model.t, rule=inverter_cap2_eq)

def daymod_eq(model, t):
    if t in day:#model.p_pvmppt[pv,t]!=0:
        return model.pg[pv,t]>=-model.sigmag[pv,t]*kdown
    return Constraint.Skip
model.daymod = Constraint(model.t, rule=daymod_eq)

def inverter_capl1_eq(model, t):
    if t in night:#model.p_pvmppt[pv,t]==0:
        return (model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.inv_cap[pv]+(model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.qg[pv,t]-model.pg[pv,t]>=model.sigmag[pv,t]*kup
    return Constraint.Skip
model.inverter_capl1 = Constraint(model.t, rule=inverter_capl1_eq)

def inverter_capl2_eq(model, t):
    if t in night:#model.p_pvmppt[pv,t]==0:
        return -(model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.inv_cap[pv]-(model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.qg[pv,t]-model.pg[pv,t]<=model.sigmag[pv,t]*kdown
    return Constraint.Skip
model.inverter_capl2 = Constraint(model.t, rule=inverter_capl2_eq)

def inverter_capl3_eq(model, t):
    if t in night:#model.p_pvmppt[pv,t]==0:
        return (model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.inv_cap[pv]-(model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.qg[pv,t]-model.pg[pv,t]>=model.sigmag[pv,t]*kup
    return Constraint.Skip
model.inverter_capl3 = Constraint(model.t, rule=inverter_capl3_eq)

def inverter_capl4_eq(model, t):
    if t in night:#model.p_pvmppt[pv,t]==0:
        return -(model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.inv_cap[pv]+(model.pnight[pv]/(model.inv_cap[pv]-model.qnight[pv]))*model.qg[pv,t]-model.pg[pv,t]<=model.sigmag[pv,t]*kdown
    return Constraint.Skip
model.inverter_capl4 = Constraint(model.t, rule=inverter_capl4_eq)

def inverter_capl5_eq(model, t):
    if t in night:#model.p_pvmppt[pv,t]==0:
        return model.pnight[pv]-model.pg[pv,t]>=model.sigmag[pv,t]*kup
    return Constraint.Skip
model.inverter_capl5 = Constraint(model.t, rule=inverter_capl5_eq)

def inverter_capl6_eq(model, t):
    if t in night:#model.p_pvmppt[pv,t]==0:
        return -model.pnight[pv]-model.pg[pv,t]<=model.sigmag[pv,t]*kdown
    return Constraint.Skip
model.inverter_capl6 = Constraint(model.t, rule=inverter_capl6_eq)

def pslack_eq(model, t):
    return model.pg[slack,t] >= 0
model.pslack = Constraint(model.t, rule=pslack_eq)

def qslack_eq(model, t):
    return model.qg[slack,t] >= 0
model.qslack = Constraint(model.t, rule=qslack_eq)

def l_eq(model, t):
    return model.loss[t] == model.pg[slack,t] + model.pg[pv,t] - sum((model.Pl[i,t])for i in model.i)
model.l = Constraint(model.t, rule=l_eq)

def pfmaxday_eq(model, t):
    if t in day:#model.p_pvmppt[pv,t]!=0:
        return model.qg[pv,t] <= model.qday[pv]
    return Constraint.Skip
model.pfmaxday = Constraint(model.t, rule=pfmaxday_eq)

def pfminday_eq(model, t):
    if t in day:#model.p_pvmppt[pv,t]!=0:
        return model.qg[pv,t] >= -model.qday[pv]
    return Constraint.Skip
model.pfminday = Constraint(model.t, rule=pfminday_eq)

def solargenlim_eq(model, t):
    return model.pg[pv,t] == model.p_pvmppt[pv,t]
model.solargenlim = Constraint(model.t, rule=solargenlim_eq)

print('Equations declared')
#OBJECTIVE--------------------------------------------------------------------------------------------------------------------
def objective_function(model):
    return sum(model.y[i,t] for i in model.i for t in model.t)
model.objective = Objective(rule=objective_function, sense=minimize, doc='Minimize voltage difference')
print('Objective declared')

print('Done constructing model!')
def pyomo_postprocess(options=None, instance=None, results=None):
    print('Pyomo printing results to file...')
    model.pprint(filename='results/increased_INV50prs_eps/results.txt')
if __name__ == '__main__':
    # This emulates what the pyomo command-line tools does
    from pyomo.opt import SolverFactory
    import pyomo.environ
    opt = SolverFactory("gurobi")
    results = opt.solve(model)#, tee=True)
    #sends results to stdout
    results.write()
    #print("\nDisplaying Solution\n" + '-'*60)
    #pyomo_postprocess(None, model, results)

print('Postprocessing results')

V_max = np.zeros((syslen,timestop))
V_min = np.zeros((syslen,timestop))

for t in model.t:
    for i in model.i:
        
        V_max[i-1,t-1] = v_max - model.sigmav[i,t]*kup
        V_min[i-1,t-1] = v_min - model.sigmav[i,t]*kdown
        

i = [ j for j in range(1,syslen+1)]

print('Plotting results...')
plt.rcParams["figure.figsize"] = (16,9)
plt.rcParams["axes.labelsize"] = 'x-large'

#GJENGANGERE------------------------------------------------------------------
#plt.plot(i,v[:,19],label='Voltage at bus')
plt.plot(i, V_max[:,19], dashes=[6,2], label='Upper voltage constraint')
plt.plot(i, V_min[:,19], dashes=[6,2], label='Lower voltage constraint')
plt.legend(loc='best')
plt.xlabel('Bus number')
plt.ylabel('Voltage (P.U)')
plt.grid()
plt.savefig('results/new/Voltage_limits_eps1.pdf', bbox_inches='tight')
plt.close()

print('All done!')
